#
# (c) 2022 Sven Lieber
# KBR Brussels
#
#import xml.etree.ElementTree as ET
import lxml.etree as ET
import os
import json
import itertools
import enchant
import hashlib
import csv
from argparse import ArgumentParser
from tqdm import tqdm
from . import descriptive_key_utils as utils
import stdnum

NS_MARCSLIM = 'http://www.loc.gov/MARC21/slim'

ALL_NS = {'marc': NS_MARCSLIM}

# -----------------------------------------------------------------------------
def main(inputFilenames, configFilename, outputFilename):
  """This script reads an XML file in MARC slim format and extracts several fields to create a CSV file."""


  with open(configFilename, 'r') as configFile:
    config = json.load(configFile)
  
  config['counters'] = {
    'recordCounter': 0,
    'name bigger than 5 words': 0,
    'fileCounter': 0,
    'filteredRecordCounter': 0,
    'filteredRecordExceptionCounter': 0
  }

  
  #
  # Instead of loading everything to main memory, stream over the XML using iterparse
  #
  with open(outputFilename, 'w') as outFile:

    outputFields = ['authorityID', 'descriptiveKey']
    outputWriter = csv.DictWriter(outFile, fieldnames=outputFields, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
    outputWriter.writeheader()

    fileCounter = 0
    recordCounter = 0
    updateFrequency = 100
    pbar = tqdm(inputFilenames)
    pbar.total = None

    # used for namespace-agnostic extraction of XML-parsed records
    recordTag = getRecordTagName(config)

    for inputFilename in pbar:
      config['counters']['fileCounter'] += 1

      # if the XML file is huge, memory becomes an issue even while streaming because a reference to the parent is kept
      # therefore we first get the root element
      # https://stackoverflow.com/questions/12160418/why-is-lxml-etree-iterparse-eating-up-all-my-memory
      context = ET.iterparse(inputFilename, tag=recordTag)
      utils.fast_iter(
        context, # the XML context
        processRecord, # the function that is called for every found recordTag
        pbar, # the progress bar that should be updated
        config, # configuration object with counters and other data
        outputWriter, # parameter for processRecord: CSV writer for output file
        updateFrequency=updateFrequency, # after how many records the progress bar should be updated
      )
      print(config['counters'])

# -----------------------------------------------------------------------------
def processRecord(elem, config, outputWriter):

  if "recordFilter" in config:
    try:
      if not utils.passFilter(elem, config["recordFilter"]):
        config['counters']['filteredRecordCounter'] += 1
        return None
    except Exception as e:
        recordID = utils.getElementValue(elem.find(config['recordIDExpression'], ALL_NS))
        config['counters']['filteredRecordExceptionCounter'] += 1
        return None

  for authorityID, descriptiveKey in getDescriptiveKeys(elem, config):
    outputWriter.writerow({"authorityID": authorityID, "descriptiveKey": descriptiveKey })




# -----------------------------------------------------------------------------
def getValueList(elem, config, configKey):

  datePatterns = config["datePatterns"]
  valueDelimiter = ';'
  keyParts = []

  # first check if we can extract the data we should extract
  #
  if configKey not in config:
    print(f'No key "{configKey}" in config!')
    return None

  # for names with many components computing the permutations takes too long
  # thus we handle a maximum of 5 permutatons
  defaultMaxPermutations = 5

  authorityID = utils.getElementValue(elem.find(config['recordIDExpression'], ALL_NS))
  #print(f'authorityID: {authorityID}')
  # check each config entry
  #
  for p in config[configKey]:
    expression = p['expression']
    values = None

    maxPermutations = p["maxPermutations"] if 'maxPermutations' in p else defaultMaxPermutations

    # extract the data
    #
    values = elem.xpath(expression, namespaces=ALL_NS)

    # process all extracted data (possibly more than one value)
    #
    if values:
      newValues = set()
      for v in values:
        vText = v.text
        vNorm = None

        if vText:
          # parse different value types, for example dates or regular strings
          #
          if 'valueType' in p:
            valueType = p['valueType']
            if valueType == 'date':
              utils.handleDate(vText, p, newValues, config['datePatterns'])
            elif valueType == 'isniURL':
              isniComponents = vText.split('isni.org/isni/')
              if len(isniComponents) > 1:
                vNorm = isniComponents[1]
                utils.addDescriptiveKeyValue(vNorm, p, newValues)
              else:
                print(f'Warning: malformed ISNI URL for authority {authorityID}: "{vText}"')
            elif valueType == 'bnfURL':
              bnfComponents = vText.split('ark:/12148/')
              if len(bnfComponents) > 1:
                vNorm = bnfComponents[1]
                utils.addDescriptiveKeyValue(vNorm, p, newValues)
              else:
                print(f'Warning: malformed BnF URL for authority {authorityID}: "{vText}"')

            else:
              # if not date or isniURL or bnfURL, parse it as text
              utils.handleText(vText, p, config, newValues, authorityID, maxPermutations)
          else:
            # parse as text by default
            utils.handleText(vText, p, config, newValues, authorityID, maxPermutations)

      keyParts.extend(newValues)
  return keyParts

# -----------------------------------------------------------------------------
def getDescriptiveKeys(elem, config):
  """This function extracts descriptive keys from the provided elements based on the provided config."""

  authorityID = utils.getElementValue(elem.find(config['recordIDExpression'], ALL_NS))
  prefixedAuthorityID = config['recordIDPrefix'] + authorityID

  # get part1's
  firstKeyParts = getValueList(elem, config, "part1")
  
  # get part2's
  secondKeyParts = getValueList(elem, config, "part2")

  # for each part1 create a combination with a part2
  descriptiveKeys = []
  for key1 in firstKeyParts:
    for key2 in secondKeyParts:
      yield [prefixedAuthorityID, f'{key1}/{key2}']

  # some key values are already unique identifiers
  uniqueKeys = getValueList(elem, config, "singlePart")
  for uk in uniqueKeys:
    yield [prefixedAuthorityID, f'{uk}']

# -----------------------------------------------------------------------------
def getRecordTagName(config):

  recordTagString = config['recordTag']
  recordTag = None
  if ':' in recordTagString:
    prefix, tagName = recordTagString.split(':')
    recordTag = ET.QName(ALL_NS[prefix], tagName)
  else:
    recordTag = recordTagString

  return recordTag



# -----------------------------------------------------------------------------
def parseArguments():

  parser = ArgumentParser(description='This script reads one or more XML files and based on a config creates descriptive keys of available field values')
  parser.add_argument('inputFiles', nargs='+', help='The inputs file containing XML records')
  parser.add_argument('-c', '--config-file', action='store', required=True, help='The config file with XPath expressions to extract')
  parser.add_argument('-o', '--output-file', action='store', required=True, help='The output CSV file containing possible descriptive keys based on the key composition config')
  args = parser.parse_args()

  return args

if __name__ == '__main__':
  args = parseArguments()
  main(args.inputFiles, args.config_file, args.output_file)



#
# (c) 2023 Sven Lieber
# KBR Brussels
#

import csv
import argparse
import work_set_clustering.lib as lib
import uuid
import time

# -----------------------------------------------------------------------------
def clusterFromScratch(inputFilenames, outputFilename, idColumnName, keyColumnName, delimiter):
  """This script performs a clustering of the input data based on common descriptive keys."""

  with open(outputFilename, 'w') as outFile:

    elementIDs = set()
    descriptiveKeys = {}

    for inputFilename in inputFilenames:
      with open(inputFilename, 'r') as inFile: 
        inputReader = csv.DictReader(inFile, delimiter=delimiter)
        lib.checkIfColumnsExist(inputReader.fieldnames, [idColumnName, keyColumnName])


        # Populate elementIDs and descriptiveKeys with values from the input file
        # A sorted list of element identifiers is returned based on the given elementIDs set
        currentElementIDs = readElements(inputReader, elementIDs, descriptiveKeys, idColumnName, keyColumnName)

        # update the list of all elements with the ones from the current file
        elementIDs.update(currentElementIDs)

    keysToElements = createInvertedIndexAndLogTime(descriptiveKeys)

    # for initial clustering these two should be empty
    clusters = {}
    elementToCluster = {}

    clusterInvertedIndex(elementIDs, descriptiveKeys, clusters, elementToCluster, keysToElements, outFile)

# -----------------------------------------------------------------------------
def updateClusters(inputFilenames, outputFilename, idColumnName, keyColumnName, delimiter, existingClustersFilename, existingClusterKeysFilename=None):
  """This script performs a clustering of the input data based on common descriptive keys."""

  with open(existingClustersFilename, 'r') as existingClustersFile, \
       open(outputFilename, 'w') as outFile:

    elementIDs = set()
    descriptiveKeys = {}

    for inputFilename in inputFilenames:
      with open(inputFilename, 'r') as inFile: 
        inputReader = csv.DictReader(inFile, delimiter=delimiter)
        lib.checkIfColumnsExist(inputReader.fieldnames, [idColumnName, keyColumnName])

        # Populate elementIDs and descriptiveKeys with values from the input file
        # A sorted list of element identifiers is returned based on the given elementIDs set
        currentElementIDs = set(readElements(inputReader, elementIDs, descriptiveKeys, idColumnName, keyColumnName))

        # update the list of all elements with the ones from the current file
        elementIDs.update(currentElementIDs)

    existingClustersReader = csv.DictReader(existingClustersFile, delimiter=delimiter)
    lib.checkIfColumnsExist(existingClustersReader.fieldnames, ['elementID','clusterID'])

    # Process optional existing descriptive keys
    if existingClusterKeysFilename:
      with open(existingClusterKeysFilename, 'r') as existingClusterKeysFile:
        existingClusterKeysReader = csv.DictReader(existingClusterKeysFile, delimiter=delimiter)
        lib.checkIfColumnsExist(existingClusterKeysReader.fieldnames, [idColumnName, keyColumnName])

        # Add descriptive key data from existing clusters (populating data structures elementIDs and descriptiveKeys
        allElementIDs = readElements(existingClusterKeysReader, elementIDs, descriptiveKeys, idColumnName, keyColumnName)
    else:
      # we leave the existing descriptive keys untouched and reuse the input elementIDs
      allElementIDs = elementIDs

    keysToElements = createInvertedIndexAndLogTime(descriptiveKeys)
    clusters = {}
    elementToCluster = {}
    readClusters(existingClustersReader, clusters, elementToCluster)

    clusterInvertedIndex(allElementIDs, descriptiveKeys, clusters, elementToCluster, keysToElements, outFile)




# -----------------------------------------------------------------------------
def createInvertedIndexAndLogTime(descriptiveKeys):
  start_time_index = time.time()
  keysToElements = lib.buildInvertedIndex(descriptiveKeys)
  end_time_index = time.time()
  diffTimeIndex = time.strftime('%H:%M:%S', time.gmtime(end_time_index - start_time_index))
  print(f'Inverted index computed in {diffTimeIndex}')
  return keysToElements

# -----------------------------------------------------------------------------
def clusterInvertedIndex(elementIDs, descriptiveKeys, clusters, elementToCluster, keysToElements, outFile):

  start_time_clustering = time.time()

  for dKey, elementIDs in keysToElements.items():

    # two helper data structures for the current iteration
    existingClusters = set()
    elementsInNoCluster = set()

    for element in elementIDs:
      if element in elementToCluster:
        existingClusters.add(elementToCluster[element])
      else:
        # those have to be added to a cluster
        elementsInNoCluster.add(element)

    if len(existingClusters) == 1:
      # sets have unique members: one or more of the elements are in the same cluster
      if len(elementsInNoCluster) > 0:
        # some of the elements are not yet in the cluster
        clusterID = existingClusters.pop()
        lib.addElementsToCluster(elementsInNoCluster, clusterID, clusters, elementToCluster)
    else:
      newClusterID = str(uuid.uuid4())

      if len(existingClusters) == 0:
        # no existing clusters found, create a new one
        clusters[newClusterID] = elementIDs
        for element in elementIDs:
          elementToCluster[element] = newClusterID

        if len(elementsInNoCluster) > 0:
          lib.addElementsToCluster(elementsInNoCluster, newClusterID, clusters, elementToCluster)

      elif len(existingClusters) > 1:
        # the members are in more than one cluster, so those clusters should be merged
        # 1. get elements of all identified clusters
        combinedElements = set()
        for cluster in existingClusters:
          combinedElements.update(clusters[cluster])
        # 2. remove old clusters
        [clusters.pop(clusterID) for clusterID in existingClusters]
        # 3. add new merged cluster
        clusters[newClusterID] = combinedElements
        # 4. update elementToCluster (overwrite if exists, otherwise create)
        for element in combinedElements:
          elementToCluster[element] = newClusterID

        # 5. now that the new cluster is created we can add eventually clusterless elements
        if len(elementsInNoCluster) > 0:
          lib.addElementsToCluster(elementsInNoCluster, newClusterID, clusters, elementToCluster)

  end_time_clustering = time.time()
  diffTimeClustering = time.strftime('%H:%M:%S', time.gmtime(end_time_clustering - start_time_clustering))
  print(f'Clusters computed in {diffTimeClustering}')

  # write cluster assignment to the output file
  outputWriter = csv.DictWriter(outFile, fieldnames=['elementID', 'clusterID'])
  outputWriter.writeheader()

  for clusterID, memberIDSet in clusters.items():
    for memberID in memberIDSet:
      outputWriter.writerow({'elementID': memberID, 'clusterID': clusterID})
 

# -----------------------------------------------------------------------------
def readElements(inputReader, elementIDs, descriptiveKeys, idColumnName, keyColumnName):

  for row in inputReader:
    eID = row[idColumnName]
    key = row[keyColumnName]

    if eID != '' and key != '':
      elementIDs.add(eID)
      if eID in descriptiveKeys:
        descriptiveKeys[eID].add(key)
      else:
        descriptiveKeys[eID] = set([key])

  return sorted(elementIDs)

# -----------------------------------------------------------------------------
def readClusters(inputReader, clusters, elementToCluster):

  for row in inputReader:
    eID = row['elementID']
    cID = row['clusterID']
    if cID not in clusters:
      clusters[cID] = set()
    lib.addElementsToCluster([eID], cID, clusters, elementToCluster)


# -----------------------------------------------------------------------------
def parseArguments():

  parser = argparse.ArgumentParser()
  parser.add_argument('-i', '--input-file', action='append', required=True, help="The CSV file(s) with columns for elements and descriptive keys, one row is one element and descriptive key relationship")
  parser.add_argument('-o', '--output-file', action='store', required=True, help='The name of the output CSV file containing two columns: elementID and clusterID')
  parser.add_argument('--id-column', action='store', required=True, help='The name of the column with element identifiers')
  parser.add_argument('--key-column', action='store', required=True, help="The name of the column that contains a descriptive key")
  parser.add_argument('--delimiter', action='store', default=',', help="Optional delimiter of the input/output CSV, default is ','")
  parser.add_argument('--existing-clusters', action='store', help="Optional file with existing element-cluster mapping")
  parser.add_argument('--existing-clusters-keys', action='store', help="Optional file with element-descriptive key mapping for existing clusters mapping")
  options = parser.parse_args()

  if options.existing_clusters_keys and not options.existing_clusters:
    print()
    print(f'Descriptive keys for elements of existing clusters provided, but no file with existing cluster keys')
    exit(1)

  return options

# -----------------------------------------------------------------------------
if __name__ == '__main__':
  options = parseArguments()
  if options.existing_clusters:
    updateClusters(options.input_file, options.output_file, options.id_column, options.key_column, options.delimiter, options.existing_clusters, options.existing_clusters_keys)
  else:
    clusterFromScratch(options.input_file, options.output_file, options.id_column, options.key_column, options.delimiter)
 



from datetime import datetime
import itertools
import xml.etree.ElementTree as ET
import unicodedata as ud
import ast
import enchant
import re

# -----------------------------------------------------------------------------
def parseYear(year, patterns):
  """"This function returns a string representing a year based on the input and a list of possible patterns.

  >>> parseYear('2021', ['%Y'])
  '2021'
  >>> parseYear('2021', ['(%Y)', '%Y'])
  '2021'
  >>> parseYear('(2021)', ['%Y', '(%Y)'])
  '2021'
  """

  parsedYear = None
  for p in patterns:

    try:
      tmp = datetime.strptime(year, p).date().year
      parsedYear = str(tmp)
      break
    except ValueError:
      pass

  if parsedYear == None:
    return year
  else:
    return parsedYear

# -----------------------------------------------------------------------------
def parseDate(date, patterns):
  """"This function returns a string representing a date based on the input and a list of possible patterns.

  >>> parseDate('2021', ['%Y'])
  '2021'
  >>> parseDate('2021', ['(%Y)', '%Y'])
  '2021'
  >>> parseDate('(2021)', ['%Y', '(%Y)'])
  '2021'

  A correct date string for a correct input.
  >>> parseDate('1988-04-25', ['%Y-%m-%d'])
  '1988-04-25'

  A correct date string for dates with slash.
  >>> parseDate('25/04/1988', ['%Y-%m-%d', '%Y/%m/%d', '%Y/%m/%d', '%d/%m/%Y'])
  '1988-04-25'

  An empty value if the pattern is not found.
  >>> parseDate('25/04/1988', ['%Y-%m-%d', '%Y/%m/%d'])
  ''

  A correct date string for dates without delimiter.
  >>> parseDate('19880425', ['%Y-%m-%d', '%Y%m%d'])
  '1988-04-25'

  Only year and month are invalid.
  >>> parseDate('1988-04', ['%Y%m', '%Y-%m'])
  ''
  >>> parseDate('198804', ['%Y-%m', '%Y%m'])
  ''

  Keep year if this is the only provided information.
  >>> parseDate('1988', ['%Y-%m-%d', '%Y'])
  '1988'

  Keep year if it is in round or square brackets or has a trailing dot.
  >>> parseDate('[1988]', ['%Y', '[%Y]'])
  '1988'
  >>> parseDate('(1988)', ['(%Y)'])
  '1988'
  >>> parseDate('1988.', ['%Y', '%Y.'])
  '1988'


  """

  parsedDate = None
  for p in patterns:

    try:
      # try if the value is a year
      tmp = datetime.strptime(date, p).date()
      if len(date) == 4:
        parsedDate = str(tmp.year)
      elif len(date) > 4 and len(date) <= 7:
        if any(ele in date for ele in ['(', '[', ')', ']', '.']):
          parsedDate = str(tmp.year)
        else:
          parsedDate = ''
      else:
        parsedDate = str(tmp)
      break
    except ValueError:
      pass

  if parsedDate == None:
    return ''
  else:
    return parsedDate



# -----------------------------------------------------------------------------
def getElementValue(elem, sep=';'):
  """This function returns the value of the element if it is not None, otherwise an empty string.

  The function returns the 'text' value if there is one
  >>> class Test: text = 'hello'
  >>> obj = Test()
  >>> getElementValue(obj)
  'hello'

  It returns nothing if there is no text value
  >>> class Test: pass
  >>> obj = Test()
  >>> getElementValue(obj)
  ''

  And the function returns a semicolon separated list in case the argument is a list of objects with a 'text' attribute
  >>> class Test: text = 'hello'
  >>> obj1 = Test()
  >>> obj2 = Test()
  >>> getElementValue([obj1,obj2])
  'hello;hello'

  In case one of the list values is empty
  >>> class WithContent: text = 'hello'
  >>> class WithoutContent: text = None
  >>> obj1 = WithContent()
  >>> obj2 = WithoutContent()
  >>> getElementValue([obj1,obj2])
  'hello'
  """
  if elem is not None:
    if isinstance(elem, list):
      valueList = list()
      for e in elem:
        if hasattr(e, 'text'):
          if e.text is not None:
            valueList.append(e.text)
      return sep.join(valueList)
    else:
      if hasattr(elem, 'text'):
        return elem.text
  
  return ''


# -----------------------------------------------------------------------------
def getNormalizedString(s):
  """This function returns a normalized copy of the given string.

  >>> getNormalizedString("HeLlO")
  'hello'
  >>> getNormalizedString("judaïsme, islam, christianisme, ET sectes apparentées")
  'judaisme islam christianisme et sectes apparentees'
  >>> getNormalizedString("chamanisme, de l’Antiquité…)")
  'chamanisme de lantiquite)'

  >>> getNormalizedString("Abe Ce De ?")
  'abe ce de'
  >>> getNormalizedString("Abe Ce De !")
  'abe ce de'
  >>> getNormalizedString("Abe Ce De :")
  'abe ce de'

  >>> getNormalizedString("A. W. Bruna & zoon")
  'a w bruna & zoon'
  >>> getNormalizedString("A.W. Bruna & Zoon")
  'aw bruna & zoon'

  #>>> getNormalizedString("---")
  #''

  #>>> getNormalizedString("c----- leopard")
  #'c leopard'
  
  """
  charReplacements = {
    '.': '',
    ',': '',
    '?': '',
    '!': '',
    ':': '',
    ';': ''
  }

  # by the way: only after asci normalization the UTF character for ... becomes ...
  asciiNormalized = ud.normalize('NFKD', s).encode('ASCII', 'ignore').lower().strip().decode("utf-8")

  normalized = ''.join([charReplacements.get(char, char) for char in asciiNormalized])
  noDots = normalized.replace('...', '')
  # remove double whitespaces using trick from stackoverflow.com/questions/8270092/remove-all-whitespace-in-a-string
  return " ".join(noDots.split())
  
  
# -----------------------------------------------------------------------------
def addDescriptiveKeyValue(value, configEntry, keyValues, prefixSuffix=''):
  """This function adds the value to found values based on some rules found in configEntry.

  >>> keys = set(['existingKey'])
  >>> addDescriptiveKeyValue("myKey",{"prefix": "myPrefix"}, keys)
  >>> sorted(keys)[1]
  'myPrefix/myKey'

  >>> addDescriptiveKeyValue("myKey2",{"prefix": "myPrefix"}, keys, prefixSuffix='Extra')
  >>> sorted(keys)[2]
  'myPrefixExtra/myKey2'
 
  """
  if value != '':
    if 'prefix' in configEntry:
      prefix = configEntry["prefix"] + prefixSuffix
      value = value.replace(' ','') if prefix == 'isni' else value
      keyValues.add(f'{prefix}/{value}')
    else:
      keyValues.add(f'{value}')

# -----------------------------------------------------------------------------
def passFilter(elem, filterConfig):
  """This function checks if the given element passes the specified filter condition.
     If the expression of the filter finds several elements, all have to pass the filter.

  The filter expression equals checks for equality
  >>> filterPseudonym = {"expression":"./datafield", "condition": "equals", "value": "p"}
  >>> elem0 = ET.fromstring("<root><datafield>p</datafield></root>")
  >>> passFilter(elem0, filterPseudonym)
  True

  >>> elem1 = ET.fromstring("<root><datafield>other value</datafield></root>")
  >>> passFilter(elem1, filterPseudonym)
  False

  An exception is thrown if the filter expression is not found
  >>> elem2 = ET.fromstring("<root><otherField>other value</otherField></root>")
  >>> passFilter(elem2, filterPseudonym)
  Traceback (most recent call last):
      ...
  Exception: Element with filter criteria not found, expected ./datafield

  An exception is thrown if multiple elements where found, but not all match the filter criteria
  >>> elem3 = ET.fromstring("<root><datafield>p</datafield><datafield>o</datafield></root>")
  >>> passFilter(elem3, filterPseudonym)
  Traceback (most recent call last):
      ...
  Exception: Multiple elements found and not all of them passed the filter: ['p', 'o'], equals p

  >>> elem4 = ET.fromstring("<root><datafield>o</datafield><datafield>p</datafield></root>")
  >>> passFilter(elem4, filterPseudonym)
  Traceback (most recent call last):
      ...
  Exception: Multiple elements found and not all of them passed the filter: ['o', 'p'], equals p

  If multiple elements where found, but all match the criteria all is good
  >>> elem5 = ET.fromstring("<root><datafield>p</datafield><datafield>p</datafield></root>")
  >>> passFilter(elem5, filterPseudonym)
  True

  The filter expression exists checks if the given element exists
  >>> filterExist = {"expression":"./datafield", "condition": "exists"}
  >>> elem6 = ET.fromstring("<root><datafield>p</datafield></root>")
  >>> passFilter(elem6, filterExist)
  True

  >>> elem7 = ET.fromstring("<root><otherField>p</otherField></root>")
  >>> passFilter(elem7, filterExist)
  False
  """

  filterExpression = filterConfig["expression"]
  condition = filterConfig["condition"]

  values = elem.xpath(filterExpression, namespaces=ALL_NS)
  if condition == "exists" or condition == "exist":
    if values:
      return True
    else:
      return False
  else:
    if values:
      filterPassed = []
      foundValues = []
      for value in values:
        foundValues.append(value.text)
        if condition == "equals" or condition == "equal":
          expectedValue = filterConfig["value"]
          if value.text == expectedValue:
            filterPassed.append(True)
          else:
            filterPassed.append(False)
          
      if all(filterPassed):
        return True
      else:
        if len(filterPassed) > 1:
          raise Exception(f'Multiple elements found and not all of them passed the filter: {foundValues}, {condition} {expectedValue}')
        else:
          return filterPassed[0]
    else:
      raise Exception(f'Element with filter criteria not found, expected {filterExpression}')

# -----------------------------------------------------------------------------
def fast_iter(context, func, pbar, config, outputWriter, updateFrequency=100, *args, **kwargs):
  """
  Adapted from http://stackoverflow.com/questions/12160418

  This function calls "func" for each parsed record in context.
  All name parameters of this function are used to initialize and update a progress bar.
  Other non-keyword arguments (args) and keyword arguments (kwargs) are provided to "func".
  """

  # We assume that context is configured to only fire 'end' events for tagName
  #
  for event, record in context:

    # call the given function and provide it the given parameters
    func(record, config, outputWriter, *args, **kwargs)

    # Update progress bar
    config['counters']['recordCounter'] += 1

    # clear to save RAM
    record.clear()
    # delete preceding siblings to save memory (https://lxml.de/3.2/parsing.html)
    while record.getprevious() is not None:
      del record.getparent()[0]

    if config['counters']['recordCounter'] % updateFrequency == 0:
      updateProgressBar(pbar, config, updateFrequency)

  # update the remaining count after the loop has ended
  updateProgressBar(pbar, config, updateFrequency)

  # We are done
  del context

# -----------------------------------------------------------------------------
def updateProgressBar(pbar, config, updateFrequency):
  """This function updates the given progress bar based on the given update frequency."""

  message = "##### get_descriptive_keys_from_xml #####"
  if "recordFilter" in config:
    passedFilter = config['counters']['recordCounter'] - config['counters']['filteredRecordCounter']
    pbar.set_description(f'{message} files: {config["counters"]["fileCounter"]}; records total: {config["counters"]["recordCounter"]}; passed filter: {passedFilter}; not passed filter: {config["counters"]["filteredRecordCounter"]}; could not apply filter: {config["counters"]["filteredRecordExceptionCounter"]}')
  else:
    pbar.set_description(f'{message} files: {config["counters"]["fileCounter"]}; records total: {config["counters"]["recordCounter"]}')
  pbar.update(updateFrequency)

# -----------------------------------------------------------------------------
def handleDictionary(vText, columnConfig, allValues, maxPermutations):
  if 'permutations' in columnConfig and columnConfig['permutations'] == 'True':
    try:
      handlePermutations(vText.values(), ' ', columnConfig, allValues, maxPermutations)
    except ValueError as e:
      vNorm = getNormalizedString(' '.join(vText.values()))
      addDescriptiveKeyValue(vNorm, columnConfig, allValues)
      print(f'{recordID}: single value "{vNorm}" taken')
  else:
    vNorm = getNormalizedString(' '.join(vText.values()))
    addDescriptiveKeyValue(vNorm, columnConfig, allValues)


# -----------------------------------------------------------------------------
def handleText(vText, columnConfig, config, allValues, recordID, maxPermutations):
  if 'permutations' in columnConfig and columnConfig['permutations'] == 'True':
    try:
      handlePermutations(vText.split(), ' ', columnConfig, allValues, maxPermutations)
    except ValueError as e:
      vNorm = getNormalizedString(vText)
      addDescriptiveKeyValue(vNorm, columnConfig, allValues)
      if 'counters' in config and 'name-more-than-maxPermutations-words' in config['counters']:
        config['counters'][f'name-more-than-maxPermutations-words'] += 1
      print(f'{recordID}: single value "{vText}" taken')
  else:
    vNorm = getNormalizedString(vText)
    addDescriptiveKeyValue(vNorm, columnConfig, allValues)

# -----------------------------------------------------------------------------
def handleDate(vText, columnConfig, allValues, datePatterns):

  vNorm = parseDate(vText, datePatterns)
  if vNorm == '':
    # pattern was not recognized: simply output input value
    addDescriptiveKeyValue(vText, columnConfig, allValues)
  else:
    # add the actually parsed date as output
    addDescriptiveKeyValue(vNorm, columnConfig, allValues)

    # additionally try to add the year of the date as separate key
    if len(vNorm) >= 4:
      possibleYear = vNorm[0:4]
      if possibleYear.isnumeric():
        yearColumnConfig = columnConfig.copy()
        yearColumnConfig['prefix'] = yearColumnConfig['prefix'] + 'Year'
        addDescriptiveKeyValue(possibleYear, yearColumnConfig, allValues)

      
       

# -----------------------------------------------------------------------------
def handlePermutations(values, joinChar, columnConfig, allValues, maxPermutations=5):

  if len(values) > maxPermutations:
    raise ValueError(f'Warning, value length={len(values)} exceeds maxPermutations={maxPermutations}') 
  else:
    for p in itertools.permutations(values): 
      vNorm = getNormalizedString(' '.join(p))
      addDescriptiveKeyValue(vNorm, columnConfig, allValues)





# -----------------------------------------------------------------------------
if __name__ == "__main__":
  import doctest
  doctest.testmod()



import unittest
import csv

# -----------------------------------------------------------------------------
class InitialClusteringSize():
  """A class with integration tests covering the number of clusters for the initial clustering"""

  # ---------------------------------------------------------------------------
  def testCorrectNumberOfClusters(self):
    """With given cluster input, two clusters should be found"""
    numberFoundClusters = len(self.getInitialClusterData()['clusterIdentifiers'])
    numberExpectedClusters = 2
    self.assertEqual(numberFoundClusters, numberExpectedClusters, msg=f'Found {numberFoundClusters} clusters instead of {numberExpectedClusters}')

# -----------------------------------------------------------------------------
class UpdateClusteringSize():
  """A class with integration tests covering the number of clusters for the initial clustering"""

  # ---------------------------------------------------------------------------
  def testCorrectNumberOfClusters(self):
    """With given cluster input, three clusters should be found"""
    numberFoundClusters = len(self.getUpdatedClusterData()['clusterIdentifiers'])
    numberExpectedClusters = 3
    self.assertEqual(numberFoundClusters, numberExpectedClusters, msg=f'Found {numberFoundClusters} clusters instead of {numberExpectedClusters}')


# -----------------------------------------------------------------------------
class InitialElementsTogether():
  """A class with integration test cases in case only the initial clustering is performed."""

  # ---------------------------------------------------------------------------
  def testElement1And2Together(self):
    """Element e1 and e2 should be clustered together"""
    clusterE1 = self.getInitialClusterData()['elementToCluster']['e1']
    clusterE2 = self.getInitialClusterData()['elementToCluster']['e2']
    self.assertEqual(clusterE1, clusterE2, msg=f'Different clusters for e1 and e2 ({clusterE1} != {clusterE2})')

  # ---------------------------------------------------------------------------
  def testElement3And4Together(self):
    """Element e3 and e4 should be clustered together"""
    clusterE3 = self.getInitialClusterData()['elementToCluster']['e3']
    clusterE4 = self.getInitialClusterData()['elementToCluster']['e4']
    self.assertEqual(clusterE3, clusterE4, msg=f'Different clusters for e3 and e4 ({clusterE3} != {clusterE4})')

# -----------------------------------------------------------------------------
class UpdateClusteringElementsTogether():
  """A class with integration test cases for updated clusters."""

  # ---------------------------------------------------------------------------
  def testElement1And5Together(self):
    """New element e5 should be clustered together with the initial e1 and e2"""
    clusterInitial = self.getUpdatedClusterData()['elementToCluster']['e1']
    clusterNew = self.getUpdatedClusterData()['elementToCluster']['e5']
    self.assertEqual(clusterInitial, clusterNew, msg=f'Different clusters for initial e1 and updated e5 ({clusterInitial} != {clusterNew})')

  # ---------------------------------------------------------------------------
  def testElement1SameClusterIDAfterUpdate(self):
    """Element e1 should have the same clusterID after update when e5 was added"""
    clusterInitial = self.getUpdatedClusterData()['elementToCluster']['e1']
    clusterNew = "c1"
    self.assertEqual(clusterInitial, clusterNew, msg=f'Initial cluster of e1 has changed after update ({clusterInitial} != {clusterNew})')

  # ---------------------------------------------------------------------------
  def testElement5SameClusterIDAfterUpdate(self):
    """Element e5 should be in the same clusterID as the initial e1"""
    clusterID = self.getUpdatedClusterData()['elementToCluster']['e5']
    clusterExpected = "c1"
    self.assertEqual(clusterID, clusterExpected, msg=f'Element e5 was not added to the existing clusterID of e1 ({clusterID} != {clusterExpected})')

  # ---------------------------------------------------------------------------
  def testElement2And5Together(self):
    """Element e5 should be clustered together with the initial e1 and e2"""
    clusterInitial = self.getUpdatedClusterData()['elementToCluster']['e2']
    clusterNew = self.getUpdatedClusterData()['elementToCluster']['e5']
    self.assertEqual(clusterInitial, clusterNew, msg=f'Different clusters for initial e2 and updated e5 ({clusterInitial} != {clusterNew})')

  # ---------------------------------------------------------------------------
  def testElement7InNewCluster(self):
    """Element e7 should be clustered in a new cluster (no overlap with initial clusters)"""
    clusterOfE7 = self.getUpdatedClusterData()['elementToCluster']['e7']
    elementsOfCluster = self.getUpdatedClusterData()['clusterToElement'][clusterOfE7]
    self.assertEqual(len(elementsOfCluster), 1, msg=f'Other elements in the cluster of e7: {elementsOfCluster}')

# -----------------------------------------------------------------------------
class UpdateClusteringNotReusingKeysElementsTogether():
  """A class with integration test cases for updated clusters when descriptive keys are not reused and no new elements can be added to existing clusters via keys."""

  # ---------------------------------------------------------------------------
  def testElement1And2Together(self):
    """Elements e1 and e2 are in one cluster according to the user"""
    clusterInitial = self.getUpdatedClusterData()['elementToCluster']['e1']
    clusterNew = self.getUpdatedClusterData()['elementToCluster']['e2']
    self.assertEqual(clusterInitial, clusterNew, msg=f'Different clusters for e1 and e2 ({clusterInitial} != {clusterNew})')

  # ---------------------------------------------------------------------------
  def testElement3Alone(self):
    """Element e3 should have its own cluster, even though there is an overlap via descriptive keys with e4"""
    clusterElements = self.getUpdatedClusterData()['clusterToElement']['c2']
    expectedNumber = 1
    self.assertEqual(len(clusterElements), expectedNumber, msg=f'Element e3 is not alone in its cluster {clusterElements}')

  # ---------------------------------------------------------------------------
  def testElement3InC2(self):
    """Element e3 should be in the cluster with identifier c3"""
    clusterInitial = self.getUpdatedClusterData()['elementToCluster']['e3']
    clusterNew = "c2"
    self.assertEqual(clusterInitial, clusterNew, msg=f'Cluster of e3 is not {clusterNew} ({clusterInitial} != {clusterNew})')

  # ---------------------------------------------------------------------------
  def testElement4Alone(self):
    """Element e4 should have its own cluster, even though there is an overlap via descriptive keys with e4"""
    clusterElements = self.getUpdatedClusterData()['clusterToElement']['c3']
    expectedNumber = 1
    self.assertEqual(len(clusterElements), expectedNumber, msg=f'Element e4 is not alone in its cluster {clusterElements}')

  # ---------------------------------------------------------------------------
  def testElement4InC3(self):
    """Element e4 should be in the cluster with identifier c3"""
    clusterInitial = self.getUpdatedClusterData()['elementToCluster']['e4']
    clusterNew = "c3"
    self.assertEqual(clusterInitial, clusterNew, msg=f'Cluster of e4 is not {clusterNew} ({clusterInitial} != {clusterNew})')

  # ---------------------------------------------------------------------------
  def testElement5Alone(self):
    """New element e5 should have its own cluster, even though there is an overlap via descriptive keys with existing elements"""
    data = self.getUpdatedClusterData()
    clusterElements = data['clusterToElement'][data['elementToCluster']['e5']]
    expectedNumber = 1
    self.assertEqual(len(clusterElements), expectedNumber, msg=f'Element e4 is not alone in its cluster {clusterElements}')

  # ---------------------------------------------------------------------------
  def testElement5InRandomClusterID(self):
    """Element e5 should be in the cluster with a new random identifier"""
    clusterID = self.getUpdatedClusterData()['elementToCluster']['e5']
    knownClusterIDs = ['c1', 'c2', 'c3']
    self.assertTrue(clusterID not in knownClusterIDs, msg=f'Cluster of e5 should have a new random ID, but it is an existing clusterID ({clusterID})')

  # ---------------------------------------------------------------------------
  def testElement6Alone(self):
    """New element e6 should have its own cluster, even though there is an overlap via descriptive keys with existing elements"""
    data = self.getUpdatedClusterData()
    clusterElements = data['clusterToElement'][data['elementToCluster']['e6']]
    expectedNumber = 1
    self.assertEqual(len(clusterElements), expectedNumber, msg=f'Element e6 is not alone in its cluster {clusterElements}')

  # ---------------------------------------------------------------------------
  def testElement6InRandomClusterID(self):
    """Element e6 should be in the cluster with a new random identifier"""
    clusterID = self.getUpdatedClusterData()['elementToCluster']['e6']
    knownClusterIDs = ['c1', 'c2', 'c3']
    self.assertTrue(clusterID not in knownClusterIDs, msg=f'Cluster of e6 should have a new random ID, but it is an existing clusterID ({clusterID})')

  # ---------------------------------------------------------------------------
  def testElement7Alone(self):
    """New element e7 should have its own cluster, even though there is an overlap via descriptive keys with existing elements"""
    data = self.getUpdatedClusterData()
    clusterElements = data['clusterToElement'][data['elementToCluster']['e7']]
    expectedNumber = 1
    self.assertEqual(len(clusterElements), expectedNumber, msg=f'Element e7 is not alone in its cluster {clusterElements}')

  # ---------------------------------------------------------------------------
  def testElement7InRandomClusterID(self):
    """Element e7 should be in the cluster with a new random identifier"""
    clusterID = self.getUpdatedClusterData()['elementToCluster']['e7']
    knownClusterIDs = ['c1', 'c2', 'c3']
    self.assertTrue(clusterID not in knownClusterIDs, msg=f'Cluster of e7 should have a new random ID, but it is an existing clusterID ({clusterID})')





#
# (c) 2024 Sven Lieber
# KBR Brussels
#
import ast
import os
import json
import enchant
import hashlib
import csv
from argparse import ArgumentParser
from tqdm import tqdm
from . import descriptive_key_utils as utils
import stdnum


# -----------------------------------------------------------------------------
def main(inputFilenames, configFilename, outputFilename):

  # read the config file
  #
  with open(configFilename, 'r') as configFile:
    config = json.load(configFile)  
  
  with open(outputFilename, 'w') as outFile:

    outputFields = ['authorityID', 'descriptiveKey']
    outputWriter = csv.DictWriter(outFile, fieldnames=outputFields, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
    outputWriter.writeheader()

    fileCounter = 0
    recordCounter = 0
    updateFrequency = 100
    pbar = tqdm(inputFilenames)
    pbar.total = None

    for inputFilename in pbar:
      fileCounter += 1

      with open(inputFilename, 'r') as inFile:
        inputReader = csv.DictReader(inFile)

        for row in inputReader:
          for authorityID, descriptiveKey in getDescriptiveKeys(row, config):
            outputWriter.writerow({"authorityID": authorityID, "descriptiveKey": descriptiveKey })

          recordCounter += 1
          if recordCounter % updateFrequency == 1:
            pbar.set_description(f'files {fileCounter}/{len(pbar)}; records: {recordCounter}')
            pbar.update()
    pbar.close()

# -----------------------------------------------------------------------------
def getDescriptiveKeys(row, config):
  """This function extracts descriptive keys from the provided row based on the provided config."""

  authorityID = row[config["recordIDColumn"]]

  # get part1's
  firstKeyParts = getValueList(row, config, "part1")
  
  # get part2's
  secondKeyParts = getValueList(row, config, "part2")

  #print(f'firstKeyParts: {firstKeyParts}')
  #print(f'secondKeyParts: {secondKeyParts}')
  #print()
  # for each part1 create a combination with a part2
  descriptiveKeys = []
  for key1 in firstKeyParts:
    for key2 in secondKeyParts:
      yield [authorityID, f'{key1}/{key2}']

  # some key values are already unique identifiers
  uniqueKeys = getValueList(row, config, "singlePart")
  for uk in uniqueKeys:
    yield [authorityID, f'{uk}']


# -----------------------------------------------------------------------------
def getValueList(row, config, configKey):

  valueDelimiter = ';'
  keyParts = []

  # first check if we can extract the data we should extract
  #
  if configKey not in config:
    print(f'No key "{configKey}" in config!')
    return None

  recordID = row[config['recordIDColumn']]

  # for names with many components computing the permutations takes too long
  # thus we handle a maximum of 5 permutatons
  defaultMaxPermutations = 5

  # check each config entry and try to add values to allValues
  #
  allValues = set()
  for columnConfig in config[configKey]:
    columnName = columnConfig['columnName']
    dataType = columnConfig['dataType'] if 'dataType' in columnConfig else 'text'

    if columnName in row:
      maxPermutations = columnConfig["maxPermutations"] if 'maxPermutations' in columnConfig else defaultMaxPermutations
      if row[columnName] != '':

        # we handle CSV files where each value is a Python list, e.g. "['value1', 'value2']" or "[{'val1': 'value1'}, {'val1': 'value1'}]"
        # get the values of the array (strings or dictionaries)
        values = ast.literal_eval(row[columnName])

        # process each element of the list of values stored in the column
        for vText in values:

          if isinstance(vText, dict):
            utils.handleDictionary(vText, columnConfig, allValues, maxPermutations)             
            
          elif isinstance(vText, str):
            if dataType == 'text':
              utils.handleText(vText, columnConfig, config, allValues, recordID, maxPermutations)

            elif dataType == 'date':
              utils.handleDate(vText, columnConfig, allValues, config['datePatterns'])

          else:
            # so far we only handle strings or dictionaries
            print(f'Script handles strings or dictionaries, but other type found for "{vText}"')

  return allValues


# -----------------------------------------------------------------------------
def parseArguments():

  parser = ArgumentParser(description='This script reads one or more CSV files and based on a config creates descriptive keys of available column values')
  parser.add_argument('inputFiles', nargs='+', help='The input file containing CSV records')
  parser.add_argument('-c', '--config-file', action='store', required=True, help='The config file with instructions how to build the descriptive keys')
  parser.add_argument('-o', '--output-file', action='store', required=True, help='The output CSV file containing descriptive keys based on the key composition config')
  args = parser.parse_args()

  return args


if __name__ == '__main__':
  args = parseArguments()
  main(args.inputFiles, args.config_file, args.output_file)



import csv
import unittest
import os
import tempfile
from work_set_clustering.clustering import clusterFromScratch as initialClustering
from work_set_clustering.clustering import updateClusters as updateClusters
from test.test_cases import InitialClusteringSize, InitialElementsTogether, UpdateClusteringSize, UpdateClusteringElementsTogether, UpdateClusteringNotReusingKeysElementsTogether

# Don't show the traceback of an AssertionError, because the AssertionError already says what the issue is!
__unittest = True

# ---------------------------------------------------------------------------
def readOutput(filename):
  """This helper function reads the output of a clusterin run and creates a data structure with cluster assignments."""

  with open(filename, 'r') as fileIn:
    csvReader = csv.DictReader(fileIn, delimiter=',')

    data = {
      'clusterIdentifiers': set(),
      'elementIdentifiers': set(),
      'elementToCluster': {},
      'clusterToElement': {}
    }

    for row in csvReader:
      elementID = row['elementID']
      clusterID = row['clusterID']
      data['elementIdentifiers'].add(elementID)
      data['clusterIdentifiers'].add(clusterID)
      data['elementToCluster'][elementID] = clusterID
      if clusterID in data['clusterToElement']:
        data['clusterToElement'][clusterID].add(elementID)
      else:
        data['clusterToElement'][clusterID] = set([elementID])

    return data

# -----------------------------------------------------------------------------
class TestClusteringSingleInput(InitialClusteringSize, InitialElementsTogether, unittest.TestCase):
  """A concrete integration test class that executes tests of the TestClustering class for a clustering with a single input files."""

  # ---------------------------------------------------------------------------
  def getInitialClusterData(self):
    return TestClusteringSingleInput.initialClusterData

  # ---------------------------------------------------------------------------
  @classmethod
  def setUpClass(cls):
    cls.tempInitialClusters = os.path.join(tempfile.gettempdir(), 'initial-clusters.csv')

    # Cluster from scratch
    #
    initialClustering(
      inputFilenames=["test/resources/cluster-input-1.csv"],
      outputFilename=cls.tempInitialClusters,
      idColumnName="elementID",
      keyColumnName="descriptiveKey",
      delimiter=","
    )

    # read the script output into an internal data structure
    #
    cls.initialClusterData = readOutput(cls.tempInitialClusters)

   # ---------------------------------------------------------------------------
  @classmethod
  def tearDownClass(cls):
    if os.path.isfile(cls.tempInitialClusters):
      os.remove(cls.tempInitialClusters)


# -----------------------------------------------------------------------------
class TestUpdateClusteringSingleInput(UpdateClusteringSize,UpdateClusteringElementsTogether, unittest.TestCase):
  """A concrete integration test class that executes tests for the update of clusters with a single input files."""

  # ---------------------------------------------------------------------------
  def getUpdatedClusterData(self):
    return TestUpdateClusteringSingleInput.updatedClusterData

  # ---------------------------------------------------------------------------
  @classmethod
  def setUpClass(cls):
    cls.tempNewClusters = os.path.join(tempfile.gettempdir(), 'updated-clusters.csv')

    # Cluster
    #
    updateClusters(
      inputFilenames=["test/resources/cluster-input-2.csv"],
      outputFilename=cls.tempNewClusters,
      idColumnName="elementID",
      keyColumnName="descriptiveKey",
      delimiter=",",
      existingClustersFilename="test/resources/clusters-1.csv",
      existingClusterKeysFilename="test/resources/cluster-input-1.csv"
    )

    # read the script output into an internal data structure
    #
    cls.updatedClusterData = readOutput(cls.tempNewClusters)

   # ---------------------------------------------------------------------------
  @classmethod
  def tearDownClass(cls):
    if os.path.isfile(cls.tempNewClusters):
      os.remove(cls.tempNewClusters)



# -----------------------------------------------------------------------------
class TestClusteringMultipleInput(InitialClusteringSize, InitialElementsTogether, unittest.TestCase):
  """A concrete integration test class that executes tests for a clustering with several input files."""

  # ---------------------------------------------------------------------------
  def getInitialClusterData(self):
    return TestClusteringMultipleInput.initialClusterData

  # ---------------------------------------------------------------------------
  @classmethod
  def setUpClass(cls):
    cls.tempInitialClusters = os.path.join(tempfile.gettempdir(), 'initial-clusters-multiple-input-files.csv')

    # Cluster from scratch
    #
    initialClustering(
      inputFilenames=["test/resources/cluster-input-1.1.csv", "test/resources/cluster-input-1.2.csv"],
      outputFilename=cls.tempInitialClusters,
      idColumnName="elementID",
      keyColumnName="descriptiveKey",
      delimiter=","
    )

    # read the script output into an internal data structure
    #
    cls.initialClusterData = readOutput(cls.tempInitialClusters)

   # ---------------------------------------------------------------------------
  @classmethod
  def tearDownClass(cls):
    if os.path.isfile(cls.tempInitialClusters):
      os.remove(cls.tempInitialClusters)

# -----------------------------------------------------------------------------
class TestClusteringOverlappingKeysDifferentClusters(UpdateClusteringNotReusingKeysElementsTogether, unittest.TestCase):
  """A concrete integration test class that executes tests of the TestClustering class for a clustering with a single input file an."""

  # ---------------------------------------------------------------------------
  def getUpdatedClusterData(self):
    return TestClusteringOverlappingKeysDifferentClusters.updatedClusterData

  # ---------------------------------------------------------------------------
  def testCorrectNumberOfClusters(self):
    """With given cluster input, six clusters should be found"""
    numberFoundClusters = len(self.getUpdatedClusterData()['clusterIdentifiers'])
    numberExpectedClusters = 6
    self.assertEqual(numberFoundClusters, numberExpectedClusters, msg=f'Found {numberFoundClusters} clusters instead of {numberExpectedClusters}')



  # ---------------------------------------------------------------------------
  @classmethod
  def setUpClass(cls):
    cls.tempNewClusters = os.path.join(tempfile.gettempdir(), 'updated-clusters-multiple-input-files.csv')

    # Cluster more
    #
    updateClusters(
      inputFilenames=["test/resources/cluster-input-2.csv"],
      outputFilename=cls.tempNewClusters,
      idColumnName="elementID",
      keyColumnName="descriptiveKey",
      delimiter=",",
      existingClustersFilename="test/resources/clusters-1-overlap.csv"
    )

    # read the script output into an internal data structure
    #
    cls.updatedClusterData = readOutput(cls.tempNewClusters)

   # ---------------------------------------------------------------------------
  @classmethod
  def tearDownClass(cls):
    if os.path.isfile(cls.tempNewClusters):
      os.remove(cls.tempNewClusters)


# -----------------------------------------------------------------------------
class TestClusteringOverlappingKeysDifferentClustersReusingKeys(UpdateClusteringSize, UpdateClusteringElementsTogether, unittest.TestCase):
  """A concrete integration test class that executes tests of the TestClustering class for a clustering with a single input file an."""

  # ---------------------------------------------------------------------------
  def getUpdatedClusterData(self):
    return TestClusteringOverlappingKeysDifferentClustersReusingKeys.updatedClusterData

  # ---------------------------------------------------------------------------
  @classmethod
  def setUpClass(cls):
    cls.tempNewClusters = os.path.join(tempfile.gettempdir(), 'updated-clusters-multiple-input-files.csv')

    # Cluster more
    #
    updateClusters(
      inputFilenames=["test/resources/cluster-input-2.csv"],
      outputFilename=cls.tempNewClusters,
      idColumnName="elementID",
      keyColumnName="descriptiveKey",
      delimiter=",",
      existingClustersFilename="test/resources/clusters-1-overlap.csv",
      existingClusterKeysFilename="test/resources/cluster-input-1.csv"
    )

    # read the script output into an internal data structure
    #
    cls.updatedClusterData = readOutput(cls.tempNewClusters)
    print(cls.updatedClusterData)

   # ---------------------------------------------------------------------------
  @classmethod
  def tearDownClass(cls):
    if os.path.isfile(cls.tempNewClusters):
      os.remove(cls.tempNewClusters)





if __name__ == '__main__':
  unittest.main()



# -----------------------------------------------------------------------------
def checkIfColumnsExist(inputColumnNames, outputColumnNames):
    """This function checks if all names of the second list are present in the first, if not an Error is raised.

    The function simply returns true if all names are present
    >>> checkIfColumnsExist(['a', 'b', 'c'], ['a', 'c'])
    True

    If a name is missing an Exception is thrown mentioning which names are missing
    >>> checkIfColumnsExist(['a', 'b', 'c'], ['a', 'd'])
    Traceback (most recent call last):
        ...
    Exception: The following requested column is not in the input: {'d'}
    """
    inputColumns = set(inputColumnNames)
    outputColumns = set(outputColumnNames)
    nonExistentColumns = outputColumns.difference(inputColumns)
    if len(nonExistentColumns) > 0:
      text = 'columns are' if len(nonExistentColumns) > 1 else 'column is'
      raise Exception(f'The following requested {text} not in the input: {nonExistentColumns}')
    else:
        return True

# -----------------------------------------------------------------------------
def addElementsToCluster(elements, clusterID, clusters, elementToCluster):
  """This function adds a mapping between elements and their cluster to the provided lookup data structures.

  >>> elements = set(['e3','e4'])
  >>> clusters = {'c1': set(['e1','e2'])}
  >>> elementToCluster = {'e1': 'c1', 'e2': 'c1'}
  >>> addElementsToCluster(elements, 'c1', clusters, elementToCluster)
  >>> sorted(clusters['c1'])
  ['e1', 'e2', 'e3', 'e4']
  >>> elementToCluster['e3']
  'c1'
  """
  clusters[clusterID].update(elements)
  for element in elements:
    elementToCluster[element] = clusterID

# -----------------------------------------------------------------------------
def buildInvertedIndex(descriptiveKeys):
  """This function builds the inverted index based on which the clustering is happening.

  >>> keys0 = {'e1': set(['key1','key2']), 'e2': set(['key3', 'key4']), 'e3': set(['key2'])}
  >>> index = buildInvertedIndex(keys0)
  >>> sorted(index['key1'])
  ['e1']
  >>> sorted(index['key2'])
  ['e1', 'e3']
  >>> sorted(index['key3'])
  ['e2']
  >>> sorted(index['key4'])
  ['e2']
  """

  keysToElements = {}
  for elementID, dKeys in descriptiveKeys.items():
    for dKey in dKeys:
      if dKey in keysToElements:
        keysToElements[dKey].add(elementID)
      else:
        keysToElements[dKey] = set([elementID])
  return keysToElements


# -----------------------------------------------------------------------------
if __name__ == "__main__":
  import doctest
  doctest.testmod()



import os
import setuptools

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setuptools.setup(
    name = "work_set_clustering",
    version = "0.4.1",
    url = "https://github.com/kbrbe/work-set-clustering",
    author = "Sven Lieber",
    author_email = "Sven.Lieber@kbr.be",
    description = ("A Python script to perform a clustering based on descriptive keys."),
    license = "AGPL-3.0",
    keywords = "FRBRization FRBR work-set-clustering",
    packages=setuptools.find_packages(),
    long_description_content_type = "text/markdown",
    long_description=read('README.md')
)
